package node

import (
	"bytes"
	"encoding/json"
	"flag"
	"log"
	"net"
	"os"
	"strings"
	"sync"
	"time"

	"git.autistici.org/ale/autoradio"
	"git.autistici.org/ale/autoradio/coordination/masterelection"
	"git.autistici.org/ale/autoradio/coordination/presence"
	"git.autistici.org/ale/autoradio/coordination/watcher"
	"git.autistici.org/ale/autoradio/instrumentation"
	"git.autistici.org/ale/autoradio/node/bwmonitor"
)

var (
	masterElectionTTL = flag.Int("master-election-ttl", 5, "TTL for the master election protocol (s)")
	nodeHeartbeat     = flag.Int("heartbeat", 3, "Period for the node presence heartbeat (s)")

	icecastReloadErrors = instrumentation.NewCounter("icecast.reload_errors")
	icecastReloads      = instrumentation.NewCounter("icecast.reload")
	configIndex         = instrumentation.NewGauge("config.etcd_index")
)

func trigger(c chan struct{}) {
	select {
	case c <- struct{}{}:
	default:
	}
}

// Converts an etcd key (a path) to the Icecast mount path. Removes
// mountPrefix from the beginning of the path, keeping the leading
// slash.
func keyToMount(key string) string {
	return key[len(autoradio.MountPrefix)-1:]
}

// In-memory representation of the overall configuration (basically
// just a list of the known mounts).
type clusterConfig struct {
	mounts map[string]*autoradio.Mount
	lock   sync.Mutex
}

func newClusterConfig() *clusterConfig {
	return &clusterConfig{
		mounts: make(map[string]*autoradio.Mount),
	}
}

// Delete implements the watcher.Syncable interface.
func (c *clusterConfig) Delete(key string) {
}

// TODO: remove?
func (c *clusterConfig) GetMount(name string) *autoradio.Mount {
	c.lock.Lock()
	defer c.lock.Unlock()
	return c.mounts[name]
}

func (c *clusterConfig) ListMounts() []*autoradio.Mount {
	c.lock.Lock()
	defer c.lock.Unlock()
	result := make([]*autoradio.Mount, 0, len(c.mounts))
	for _, m := range c.mounts {
		result = append(result, m)
	}
	return result
}

// Keeps the in-memory service configuration (clusterConfig) in sync
// with the etcd database. An update channel is triggered whenever the
// data changes.
type configWatcher struct {
	*clusterConfig
	client autoradio.EtcdClient
	update chan struct{}
}

func newConfigWatcher(client autoradio.EtcdClient, config *clusterConfig, update chan struct{}) *watcher.Syncer {
	return watcher.NewSyncer(client, autoradio.MountPrefix, &configWatcher{
		clusterConfig: config,
		client:        client,
		update:        update,
	})
}

// Set implements the watcher.Syncable interface. It will update the
// clusterConfig only if the data actually changed.
func (c *configWatcher) Set(key, value string, index uint64) {
	name := keyToMount(key)

	var m autoradio.Mount
	if err := json.NewDecoder(strings.NewReader(value)).Decode(&m); err != nil {
		log.Printf("error updating mount %s [@%d]: corrupted data: %v", name, index, err)
		return
	}

	c.lock.Lock()
	defer c.lock.Unlock()
	if prev, ok := c.mounts[name]; ok && prev.Equal(&m) {
		// No changes.
		return
	}

	c.mounts[name] = &m
	log.Printf("updated mount %s [@%d]: %s", name, index, value)
	// trigger update here
	trigger(c.update)
}

// Delete implements the watcher.Syncable interface.
func (c *configWatcher) Delete(key string) {
	name := keyToMount(key)

	c.lock.Lock()
	defer c.lock.Unlock()
	if _, ok := c.mounts[name]; ok {
		delete(c.mounts, name)
	}
	// trigger update here
	trigger(c.update)
}

// Private interfaces for process controllers. These are used to
// replace real processes with mocks while testing.
type controller interface {
	Update(*clusterConfig, bool, net.IP) error
	GetStatus() *icecastStatus
	Run(chan bool)
}

type transcodingController interface {
	Start()
	Stop()
}

// Factory for transcodingControllers.
type transcodingControllerFunc func(*liquidsoapParams) (transcodingController, error)

// RadioNode is an active streaming node, managing the local icecast server.
type RadioNode struct {
	wg     sync.WaitGroup
	client autoradio.EtcdClient

	name        string
	ips         []net.IP
	internalIPs []net.IP
	config      *clusterConfig

	me     *masterelection.MasterElection
	syncer *watcher.Syncer
	bw     *bwmonitor.BandwidthUsageMonitor

	icecast      controller
	maxListeners int

	// Node presence heartbeat.
	presence *presence.Presence

	// Rate limiting for Icecast daemon restarts.
	reloadDelay time.Duration

	// Generator for transcodingControllers. Exposed as a member
	// so that it can be stubbed out during tests.
	transcoderFn transcodingControllerFunc

	// All currently active transcoders (locked due to the
	// async debugging handler).
	transcodersMx sync.Mutex
	transcoders   map[string]*transcoder

	// A note on channel types used for signaling: while I
	// personally prefer struct{} chans, the etcd interface for
	// Watch makes it convenient to use bool stop channels
	// throughout the application.
	upch chan struct{}
	stop chan bool

	// Logger for debug messages and state changes.
	Log *log.Logger
}

// NewRadioNode creates and initializes a new autoradio node.
func NewRadioNode(name string, ips, internalIPs []net.IP, netDev string, bwLimit float64, maxListeners int, client autoradio.EtcdClient) *RadioNode {
	// Global 'stop' channel.
	stopch := make(chan bool)

	// Network updates trigger icecast reconfiguration. This
	// channel is used as an event signal.
	upch := make(chan struct{}, 1)

	// MasterElection changes trigger an update.
	mech := make(chan masterelection.State)
	go func() {
		for range mech {
			trigger(upch)
		}
	}()

	// Location information advertised when this node is master.
	minfo := &autoradio.MasterNodeInfo{
		Name:       name,
		IP:         ips,
		InternalIP: internalIPs,
	}
	minfodata, err := json.Marshal(minfo)
	if err != nil {
		log.Fatal(err)
	}

	// Create the RadioNode and all the auxiliary objects it
	// contains. Note that the per-node icecast client limit is
	// actually set to a value greater than maxListeners, to allow
	// for some headroom in the front-end traffic control
	// computations (if everything goes well, connections should
	// never be rejected by Icecast).
	config := newClusterConfig()
	rc := &RadioNode{
		config:      config,
		name:        name,
		ips:         ips,
		internalIPs: internalIPs,
		client:      client,
		me: masterelection.New(
			client,
			autoradio.MasterElectionPath,
			string(minfodata),
			uint64(*masterElectionTTL),
			mech),
		syncer:  newConfigWatcher(client, config, upch),
		icecast: newIcecastController(name, maxListeners*2),
		transcoderFn: func(p *liquidsoapParams) (transcodingController, error) {
			return newLiquidsoap(p)
		},
		transcoders:  make(map[string]*transcoder),
		reloadDelay:  1000 * time.Millisecond,
		bw:           bwmonitor.NewBandwidthUsageMonitor(netDev, bwLimit),
		maxListeners: maxListeners,
		upch:         upch,
		stop:         stopch,
		Log:          log.New(os.Stderr, "node: ", 0),
	}
	rc.presence = presence.New(client, autoradio.NodePrefix, rc.getNodeStatus, uint64(*nodeHeartbeat))
	return rc
}

func (rc *RadioNode) getNodeStatus() string {
	// Build our NodeStatus.
	icecastStatus := rc.icecast.GetStatus()
	nodeStatus := autoradio.NodeStatus{
		Name:           rc.name,
		IP:             rc.ips,
		InternalIP:     rc.internalIPs,
		IcecastUp:      icecastStatus.Up,
		Mounts:         icecastStatus.Mounts,
		BandwidthUsage: rc.bw.GetUsage(),
		MaxListeners:   rc.maxListeners,
	}

	// Update our node entry in the database.
	var buf bytes.Buffer
	json.NewEncoder(&buf).Encode(&nodeStatus)
	return buf.String()
}

// Get a valid internal IP address for the current master node (to be
// passed to Icecast). Since we don't really know much about network
// topology, just pick the first IP address associated with the master
// node.
func (rc *RadioNode) getMasterInternalAddr() net.IP {
	var info autoradio.MasterNodeInfo
	if err := json.NewDecoder(strings.NewReader(rc.me.GetMasterData())).Decode(&info); err != nil || len(info.GetInternalIP()) == 0 {
		return nil
	}
	return info.GetInternalIP()[0]
}

// Reload the icecast configuration when needed.
func (rc *RadioNode) updater(stop chan bool) {
	// Wait an instant to give a chance to the services to
	// initialize properly.
	time.Sleep(200 * time.Millisecond)

	// Keep track of all the configured transcoders (and clean
	// them up at the end).
	defer func() {
		for _, t := range rc.transcoders {
			t.Stop()
		}
	}()

	rc.Log.Printf("starting updater")
	for {
		select {
		case <-rc.upch:
			// We may have received an update before the
			// masterelection had time to actually elect a
			// master. In this case, skip.
			if !rc.me.Valid() {
				continue
			}

			rc.Log.Printf("updating configuration")

			masterAddr := rc.getMasterInternalAddr()

			// Reload the Icecast daemon.
			icecastReloads.Incr()
			if err := rc.icecast.Update(rc.config, rc.me.IsMaster(), masterAddr); err != nil {
				icecastReloadErrors.Incr()
				rc.Log.Printf("Update(): %v", err)
			}

			// Check the configuration for new or removed
			// transcoding mounts, and start (or stop) the
			// associated transcoder objects. We also need
			// to detect changes in the encoding params
			// and restart the transcoder if necessary.
			rc.transcodersMx.Lock()
			tmp := make(map[string]struct{})
			for name := range rc.transcoders {
				tmp[name] = struct{}{}
			}
			for _, m := range rc.config.ListMounts() {
				if !m.HasTranscoder() {
					continue
				}

				tparams := newLiquidsoapParams(m)
				cur, ok := rc.transcoders[m.Name]
				if ok {
					delete(tmp, m.Name)
					if cur.Changed(tparams) {
						cur.Stop()
						ok = false
					}
				}
				if !ok {
					if t, err := newTranscoder(tparams, rc.transcoderFn, rc.name, rc.client); err != nil {
						rc.Log.Printf("could not create transcoder: %v", err)
					} else {
						t.Start()
						rc.transcoders[m.Name] = t
					}
				}
			}
			for name := range tmp {
				rc.transcoders[name].Stop()
				delete(rc.transcoders, name)
			}
			rc.transcodersMx.Unlock()

			// Limit the rate of reconfigurations.
			if rc.reloadDelay > 0 {
				time.Sleep(rc.reloadDelay)
			}

		case <-stop:
			return
		}
	}
}

// Start the node. Returns immediately.
func (rc *RadioNode) Start() {
	// Starting the presence runner might fail - raise a fatal
	// error in that  case.
	if err := rc.presence.Start(); err != nil {
		log.Fatal(err)
	}

	// Start auxiliary services and event listeners on their own
	// goroutines. Put them all in a WaitGroup so we can wait for
	// their termination.
	bgfuncs := []func(chan bool){
		// Icecast updater,
		rc.updater,

		// Master election runner.
		rc.me.Run,

		// Icecast status checker.
		rc.icecast.Run,

		// Bandwidth monitor.
		rc.bw.Run,
	}

	for _, fn := range bgfuncs {
		rc.wg.Add(1)
		go func(fn func(stop chan bool)) {
			fn(rc.stop)
			rc.wg.Done()
		}(fn)
	}
}

// Wait until all processing associated with this node has terminated.
func (rc *RadioNode) Wait() {
	rc.wg.Wait()
	// Let's leave this around for garbage collection, otherwise
	// the in-memory etcd mock server might try to send values to
	// it due to a Watch-trigger desynchronization...
	//close(rc.upch)
}

// Run the node, waiting for termination.
func (rc *RadioNode) Run() {
	rc.Start()
	rc.Wait()
}

// Stop everything.
func (rc *RadioNode) Stop() {
	rc.presence.Stop()
	close(rc.stop)
}

// Transcoder just runs a master election protocol and starts
// liquidsoap for a submount whenever this node becomes the master.
// Transcoder instances can be started and stopped individually. The
// transcoding parameters set at creation time can't be changed while
// the transcoder is running (it must be stopped and restarted).
type transcoder struct {
	params     *liquidsoapParams
	nodeName   string
	client     autoradio.EtcdClient
	liquidsoap transcodingController
	stop       chan bool
	wg         sync.WaitGroup
}

func newTranscoder(params *liquidsoapParams, tfn transcodingControllerFunc, nodeName string, client autoradio.EtcdClient) (*transcoder, error) {
	l, err := tfn(params)
	if err != nil {
		return nil, err
	}
	return &transcoder{
		params:     params,
		liquidsoap: l,
		client:     client,
		nodeName:   nodeName,
		stop:       make(chan bool),
	}, nil
}

// Changed returns true if the stream parameters have changed,
// requiring a transcoder restart.
func (t *transcoder) Changed(newParams *liquidsoapParams) bool {
	return !t.params.Equal(newParams)
}

func (t *transcoder) run() {
	defer t.wg.Done()

	// The master election protocol must be stopped when the
	// transcoder terminates, so its lifecycle is fully contained
	// within the scope of this function.
	update := make(chan masterelection.State)
	mestop := make(chan bool)
	me := masterelection.New(
		t.client,
		autoradio.TranscoderMasterElectionBase+t.params.TargetMount,
		t.nodeName,
		uint64(*masterElectionTTL),
		update)
	go me.Run(mestop)
	defer close(mestop)

	running := false
	for {
		select {
		case state := <-update:
			if state.Role == masterelection.RoleMaster {
				t.liquidsoap.Start()
				running = true
			} else if running {
				t.liquidsoap.Stop()
				running = false
			}
		case <-t.stop:
			if running {
				t.liquidsoap.Stop()
			}
			return
		}
	}
}

func (t *transcoder) Start() {
	t.wg.Add(1)
	go t.run()
}

func (t *transcoder) Stop() {
	close(t.stop)
	t.wg.Wait()
}
